/* matrix.cpp                                                   10/04/00

     Copyright (c) John R. Berryhill 1995 - 2000.  All Rights Reserved.

*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h> // for mem* and cout
#include "..\book\source\vector.h"
#include "..\book\source\matrix.h"
#include "..\book\source\sparse.h"

enum{NO,YES} DEBUG = NO;

submatrix::submatrix(const matrix& A, unsigned i, unsigned j)
{
   i--; j--; mrows = A.mrows-i; ncols = A.ncols-j; size = 0;
   v = new real_vector [mrows];
   for (unsigned k=0; k<mrows; k++) { v[k].ptr = A[k+i].ptr+j;
                                      v[k].n   = ncols;       }
}

submatrix::submatrix(const real_vector& x, unsigned j)
{
   j--; mrows = x.n-j; ncols = 1; size = 0;
   v = new real_vector [mrows];
   for (unsigned k=0; k<mrows; k++) { v[k].ptr = x.ptr+j+k;
                                      v[k].n   = 1;       }
}

submatrix::~submatrix()
{                                     if (DEBUG) printf("del %Fp \n",v);
   delete v;                    // will inline
}

submatrix& submatrix::operator--()
{
   mrows--; ncols--;
   for (unsigned k=0; k<mrows; k++) --v[k];
   return *this;
}

matrix::matrix(unsigned i, unsigned j)
{
   mrows = i;  ncols = j;  size = i*j;  v = new real_vector [mrows];
   double* space = new double [size];  checkptr(space); // contiguous!
   memset(space, 0, size*sizeof(double));
   for (unsigned k=0; k<mrows; k++) v[k].init(ncols, space+k*ncols);
}

matrix::matrix(double* data, unsigned i, unsigned j)
{
   mrows = i;  ncols = j;  size = 0;  v = new real_vector [mrows];
   for (unsigned k=0; k<mrows; k++) v[k].init(ncols, data+k*ncols);
}

matrix& matrix::plus_outer (const real_vector& l, const real_vector& r)
{
   unsigned i,j;
   if (mrows != l.n || ncols != r.n) exit(1);
   for(i=0; i<mrows; i++) for(j=0; j<ncols; j++) v[i][j] += l[i] * r[j];
   return *this;
}

void matrix::swaprows (const unsigned i, const unsigned j)
{
   unsigned l; double* dat; //in lieu of real_vector* u;
          dat = v[i-1].ptr;         l = v[i-1].n;
   v[i-1].ptr = v[j-1].ptr;  v[i-1].n = v[j-1].n;
   v[j-1].ptr = dat;         v[j-1].n = l;
}

matrix& matrix::row_house(const real_vector& x)
{//G&VL 5.1.2 p197; includes stashing x
   double beta = x * x; if (beta == 0.) exit(1); beta = -2./beta;
   plus_outer( x, (x*(*this))*=beta );
   for (unsigned i=1; i<mrows; i++) v[i][0] = x[i];
   return *this;
}

matrix& matrix::col_house(const real_vector& x)
{//G&VL 5.1.3 p197; includes stashing x
   double beta = x * x; if (beta == 0.) exit(1); beta = -2./beta;
   plus_outer( ((*this)*x)*=beta, x );
   for (unsigned j=1; j<ncols; j++) v[0][j] = x[j];
   return *this;
}

matrix& matrix::HouseholderQR()
{//G&VL 5.2.1 p212
   for(unsigned i=1; i<=ncols; i++){
      submatrix C(*this,i,i);
      C.row_house( house(C,1) );    }
   return *this;
}

matrix& matrix::HouseBidiag()
{//G&VL 5.4.2 p237
   for(unsigned j=1; j<=ncols; j++){
      submatrix C(*this,j,j);
      C.row_house( house(C,1) );
      if(j > ncols-2) continue;
      submatrix D(C,1,2);
      D.col_house( house( submatrix(D[0],1), 1) ); }
   return *this;
}

matrix& matrix::HouseTridiag()
{//G&VL p420; algorithm 8.2.1: symmetric QR
   if(mrows != ncols) exit(1);
   double* space = new double [2*ncols];
   real_vector v(space,ncols), w(space+ncols,ncols);
   double beta;
   for(unsigned j=1; j<ncols-1; j++)
   {  --v; --w;
      submatrix C(*this,j,j);
      v = sym_house(C);
      beta = v * v;
      submatrix D(*this,j+1,j+1);
      w  = D * v;
      w *= (-2./beta);
      w -= (v * ((w * v) / beta));
      D.plus_outer(w,v).plus_outer(v,w);
   }
   delete space;
   return *this;
}

unsigned matrix::col_rank (double drop_tol)
{
   unsigned r = ncols;
   for(unsigned i=0; i<ncols; i++)
      if( fabs( v[i][i] ) < drop_tol ) { v[i][i] = 0.; r--; }
   return r;
}

void matrix::dumpall () const
{
   unsigned i, j;
   for (i=0; i<mrows; i++){
      cout << (i+1);
      for (j=0; j<ncols; j++) cout << "|" << v[i][j];
      cout << "\n";       }
}

unsigned Jacobi( real_vector& x, real_vector& y,
                 real_vector& p, real_vector& q )
{//Greenstadt (in R&W) p85; Compute c and s from x and y
   double e = 1.0e-12;
   double alfa = x*x; if( sqrt(alfa) <= e ) return 0;
   double beta = y*y; if( sqrt(beta) <= e ) return 0;
   double gama = x*y; if( fabs(gama) <= sqrt(alfa*beta)*e ) return 0;
   double xi = 0.5*(beta-alfa)/gama;
   double t = 1./(fabs(xi)+sqrt(1.+xi*xi));
   double c = 1./sqrt(1.+t*t);
   double s = t*c;  if( xi < 0.) s = -s;
// Apply J to x and y
   for(unsigned i=0; i<x.N(); i++){
      t = x[i]; x[i] = x[i]*c - y[i]*s; y[i] = y[i]*c + s*t; }
// gama = x*y; // just for inspection
// Apply J to p and q
   for(unsigned i=0; i<p.N(); i++){
      t = p[i]; p[i] = p[i]*c - q[i]*s; q[i] = q[i]*c + s*t; }
   return 1;
}

void SVD (matrix& A, matrix& I, real_vector& w)
{//G&VL p456, transposed
   unsigned i, j, k, m = A.M(); double norm, e = 1.0e-12;

   for(i=0; i<m; i++){
      do{k = 0;
         for(j=i+1; j<m; j++)
            k += Jacobi( A[j], A[i], I[j], I[i]);
      }while (k>0);

      norm = sqrt(A[i]*A[i]);
      if(norm > e){ w[i] = norm;  A[i] *= (1./norm); }
         else     { w[i] = 0.;    A[i].init();  I[i].init(); }
   }
   return;
}

void givens_2x2_prepost (matrix& T, unsigned k, double a, double b)
{//c and s per G&VL p202; algorithm 5.1.5:
   double c, s, r, t;
   if(b == 0.) return; //{c = 1.; s = 0.;}
     else{
       if(fabs(b) > fabs(a)){
         r = -a/b; s = 1./sqrt(1.+r*r); c = s*r;}
         else { r = -b/a; c = 1./sqrt(1.+r*r); s = c*r;}
     }
// special supporting alg 8.2.2 p.423:
   unsigned i=k-1, j=k+1, l=k+2, n=T.N();
   if(k>1){
   r = T(k,i); t = T(j,i); T(k,i) = c*r-s*t; T(j,i) = s*r+c*t;}//col0
   r = T(k,k); t = T(j,k); T(k,k) = c*r-s*t; T(j,k) = s*r+c*t; //col1
   r = T(k,j); t = T(j,j); T(k,j) = c*r-s*t; T(j,j) = s*r+c*t; //col2
// r = T(1,3); t = T(2,3); T(1,3) = c*r-s*t; T(2,3) = s*r+c*t;   col3
   if(n>j){ T(k,l) = -s*T(j,l); T(j,l) *= c;}
   if(k>1){
   r = T(i,k); t = T(i,j); T(i,k) = c*r-s*t; T(i,j) = s*r+c*t;}//row0
   r = T(k,k); t = T(k,j); T(k,k) = c*r-s*t; T(k,j) = s*r+c*t; //row1
   r = T(j,k); t = T(j,j); T(j,k) = c*r-s*t; T(j,j) = s*r+c*t; //row2
// r = T(3,1); t = T(3,2); T(3,1) = c*r-s*t; T(3,2) = s*r+c*t;   row3
   if(n>j){ T(l,k) = -s*T(l,j); T(l,j) *= c;}
   return;
}

void givens_2x2_symmetric(matrix& T, unsigned k, double a, double b)
{//exploits symmetry: neither reads from nor writes to T above diag.
   double c, s, r, t, u, v;
   if(b == 0.) return; //{c = 1.; s = 0.;}
     else{
       if(fabs(b) > fabs(a)){
         r = -a/b; s = 1./sqrt(1.+r*r); c = s*r;}
         else { r = -b/a; c = 1./sqrt(1.+r*r); s = c*r;}
     }
   unsigned i=k-1, j=k+1, l=k+2, n=T.N();
   if(k>1){
   r = T(k,i); t = T(j,i); T(k,i) = c*r-s*t; T(j,i) = s*r+c*t;}//col0
   r = T(k,k); v = T(j,k); T(k,k) = c*r-s*v; T(j,k) = s*r+c*v; //col1
               t = T(j,j);      u = c*v-s*t; T(j,j) = s*v+c*t; //col2
   r = T(k,k);             T(k,k) = c*r-s*u;                   //row1
   r = T(j,k); t = T(j,j); T(j,k) = c*r-s*t; T(j,j) = s*r+c*t; //row2
   if(n>j){ T(l,k) = -s*T(l,j); T(l,j) *= c;}
   return;
}

void alg_822( matrix& T )
{//G&VL p423 symmetric QR step with implicit shift
   double d, mu, a, b; unsigned k, n = T.N();
   d = ( T(n-1,n-1) - T(n,n) ) * .5;
   a = T(n,n-1)*T(n,n-1); mu = sqrt(d*d+a); if(d < 0.) mu = -mu;
   mu = d + mu; mu = T(n,n) - a/mu;
   a = T(1,1) - mu; b = T(2,1);
   for (k=1; k<n; k++)
   {
      givens_2x2_prepost(T, k, a, b);
      if(k > n-2) break;
      a = T(k+1,k); b = T(k+2,k);
   }
}

void alg_823( matrix& A )
{//G&VL p424 Symmetric QR Diagonalization
   unsigned i, n;  double e=1.0e-12;
   submatrix T(A,1,1);
   while ( (n=T.N()) > 1){  i = n-1;
      alg_822( T );
      if(fabs(T(i+1,i)) <= (fabs(T(i,i))+fabs(T(i+1,i+1))) * e){
         T(i+1,i) = T(i,i+1) = 0.;  --T;
      }
   }
}

unsigned Crout_LU( matrix& A, pivot& P )
{//Crout elimination column by column
   unsigned i,j,k,p;  double Ujj, Ukj;

   for (j=1; j<=A.N(); j++)  //col-by-col computation of L\U
   {
//    Recursion on the first j-1 entries of col j
      for(k=1; k<j; k++){
         if( (Ukj = A(k,j)) == 0.) continue;
//       prior column k of L:
         for(i=k+1; i<=A.M(); i++) A(i,j) -= A(i,k)*Ukj;
      }
//    Find the pivot
      Ujj = A(j,j);  p = j;
      for(i=j+1; i<=A.M(); i++){
         if(fabs(A(i,j)) <= fabs(Ujj)) continue;
         Ujj = A(i,j);  p = i;
      }
      if( Ujj == 0.) break; //return j-1;
//    Swap rows
      if(p != j){ A.swaprows(j,p);  P.swaprows(j,p); }
//    Finalize
      for(i=j+1; i<=A.M(); i++){
         if (A(i,j) == 0.) continue;
         A(i,j) /= Ujj;                     //Lij
      }
// printf("j: %u\n",j); A.dumpall(); WaitForIt();
   }
   return j-1;
}
