/* autderiv.cpp                                                 12/29/00

     Copyright (c) John R. Berryhill 2000.  All Rights Reserved.
     Self-contained archive of the automatic derivative stuff in
     Chapt 9. Declarations, implementations, tests, and main().

*/
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <iostream.h>

void WaitForIt ()
{
   char ch = '\0';
   fprintf(stderr,"Press ENTER to continue . . . ");
   while (!(ch == '\n' || ch == '\r'))
      ch = (char)getchar();
   return;
}

typedef double (*differentiable) (double);
double negsin (double x) { return -sin(x); }
double inv    (double x) { return 1. / x;  }
double dsqrt  (double x) { return 0.5 / sqrt(x); }
double square (double x) { return x * x; }
double dsquare(double x) { return 2. * x; }
double cube   (double x) { return x * x * x; }
double dcube  (double x) { return 3. * x * x; }
double datan  (double x) { return 1./(1.+x*x); }
double dasin  (double x) { return 1./sqrt((1.+x)*(1.-x)); }
double dtan   (double x) { return 1./(cos(x)*cos(x)); }
double itself (double x) { x; return x; }
double oneval (double x) { x; return 1.0; }

class C1 { public:                         //only requires () and[]
   virtual double operator()(double x) = 0;
   virtual double operator[](double x) = 0;
};

class dif1 : public C1
{
   differentiable f, df; public:

   dif1(){ f = df = NULL; }
   dif1( double(*a)(double), double(*b)(double) ){ f = a; df = b; }
   double operator()(double x){ return f(x); }
   double operator[](double x){ return df(x);}
   double deriv( double x)    { return df(x);}
   friend differentiable value (dif1& a){ return a.f;  }
   friend differentiable deriv (dif1& a){ return a.df; }
};
dif1 sinf ( sin,    cos );
dif1 cosf ( cos,    negsin );
dif1 logf ( log,    inv );
dif1 sqrtf( sqrt,   dsqrt );
dif1 sqf  ( square, dsquare );
dif1 cubf ( cube,   dcube);
dif1 expf ( exp,    exp );
dif1 tanf (  tan, dtan );
dif1 atanf( atan, datan);
dif1 asinf( asin, dasin);
dif1 arg  ( itself, oneval );

class Sum : public C1
{
   C1 *f1, *f2;        public:

   Sum(C1& a, C1& b){ f1 = &a; f2 = &b; }
   double operator()(double x){ return (*f1)(x) + (*f2)(x); }
   double operator[](double x){ return (*f1)[x] + (*f2)[x]; }
   double deriv( double x)    { return (*f1)[x] + (*f2)[x]; }
};
inline Sum operator+ (C1& a, C1& b){ return Sum(a,b); }

class Dif : public C1
{
   C1 *f1, *f2;        public:

   Dif(C1& a, C1& b){ f1 = &a; f2 = &b; }
   double operator()(double x){ return (*f1)(x) - (*f2)(x); }
   double operator[](double x){ return (*f1)[x] - (*f2)[x]; }
   double deriv( double x)    { return (*f1)[x] - (*f2)[x]; }
};
inline Dif operator- (C1& a, C1& b){ return Dif(a,b); }

class Prod : public C1
{
   C1 *f1, *f2;        public:

   Prod(C1& a, C1& b){ f1 = &a; f2 = &b; }
   double operator()(double x){ return (*f1)(x) * (*f2)(x); }
   double operator[](double x){
                  return (*f1)[x]*(*f2)(x)+(*f1)(x)*(*f2)[x]; }
};
inline Prod operator* (C1& a, C1& b){ return Prod(a,b); }

class Quot : public C1
{
   C1 *f1, *f2;        public:

   Quot(C1& a, C1& b){ f1 = &a; f2 = &b; }
   double operator()(double x){ return (*f1)(x) / (*f2)(x); }
   double operator[](double x){ return ((*f1)[x] - (*f1)(x)*(*f2)[x]/
                                        (*f2)(x)) /(*f2)(x);}
};
inline Quot operator/ (C1& a, C1& b){ return Quot(a,b); }

class Parm : public C1
{
   double c;             public:

   Parm(){ c = 0.; }
   Parm( double a ){ c = a; }
   double operator()(double x){ x; return c; }
   double operator[](double x){ x; return 0.;}
   void   operator= (double x){ c = x; }
};
inline Sum operator+ (C1& a, double b){ return a + Parm(b); }
inline Sum operator+ (double a, C1& b){ return Parm(a) + b; }
inline Dif operator- (C1& a, double b){ return a - Parm(b); }
inline Dif operator- (double a, C1& b){ return Parm(a) - b; }
inline Prod operator* (C1& a, double b){ return a * Parm(b); }
inline Prod operator* (double a, C1& b){ return Parm(a) * b; }
inline Quot operator/ (C1& a, double b){ return a / Parm(b); }
inline Quot operator/ (double a, C1& b){ return Parm(a) / b; }

class Variable : public C1{ public:
   Variable(){}
   double operator()(double x){ return x; }
   double operator[](double x){ x; return 1.;}
};

class C1_function : public C1 {
   C1 *f;           public:

   C1_function()   { f = NULL; }
   C1_function(C1& a){ f = &a; }
   double operator()(double x){ return (*f)(x); }
   double operator[](double x){ return (*f)[x]; }
   void   operator= (C1&    x){ f = &x; }
};

class Chain : public C1
{
   C1    *C1ptr;
   Chain *next;
   char  *note;  public:

   Chain(){ C1ptr = NULL; next = this; note = "";}
   Chain( C1& B ){ C1ptr = &B; next = this; note = "";}
   Chain( C1& B, char* name ){ C1ptr = &B; next = this; note = name;}

   void setC1( C1& B){ C1ptr = &B; }
   void setnote( char* name){ note = name; }
   Chain& operator >> ( Chain& a ){ next = &a;  return a; }
   friend ostream& operator << (ostream& s, Chain& q);
   double operator()(double x);
   double operator[](double x);
};

double Chain::operator () ( double x )
{
   C1 *func;  Chain *now;  now = this;
   for(;;){
      func = now->C1ptr;  x = (*func)(x);
      if( now == now->next ) break;  now = now->next;
   }
   return x;
}

double Chain::operator [] ( double x )
{
   double dx = 1.0;  C1 *func;  Chain *now;  now = this;
   for(;;){
      func = now->C1ptr;  dx *= (*func)[x];  x = (*func)(x);
      if( now == now->next ) break;  now = now->next;
   }
   return dx;
}

ostream& operator << (ostream& s, Chain& q)
{
      s << "[" << &q      << " -> "
               << q.next  << " | "
               << q.C1ptr << " | "
               << q.note  << " ]\n";
      return s;
}

void ex_9_1_1(void)
{
   differentiable z;
   dif1 sinf( sin, cos );
   dif1 cosf( cos, negsin );
   double qpi = 2.*atan(1.0)/3.; //(30deg)
   cout << "1:" << value(sinf)(qpi) << endl;
   cout << "2:" << deriv(sinf)(qpi) << endl;
   cout << "3:" << value(cosf)(qpi) << endl;
   cout << "4:" << deriv(cosf)(qpi) << endl;
   z = deriv(cosf);
   cout << "5:" << z(qpi)           << endl;

   cout << "1:" << sinf(qpi) << endl;
   cout << "2:" << sinf[qpi] << endl;
   cout << "3:" << cosf(qpi) << endl;
   cout << "4:" << cosf[qpi] << endl;
   cout << "5:" << cosf.deriv(qpi) << endl;
}

void ex_9_3_1(void)
{
   Sum poly(sqf,cubf);
   cout << "1:" << poly(2.) << endl;
   cout << "2:" << poly[2.] << endl;
   cout << "3:" << (sqf+cubf)(2.) << endl;
   cout << "4:" << (sqf+cubf)[2.] << endl;
   C1_function z;
   z = poly;
   cout << "5:" << z(3) << endl;
   cout << "6:" << z[3] << endl;
}

void ex_9_3_2(void)
{
// C1 z; illegal!
   C1 *z;
   Prod fifth (sqf,cubf);
   cout << "1:" << fifth(2.)      << endl;
   cout << "2:" << fifth[2.]      << endl;
   cout << "3:" << (sqf/cubf)(2.) << endl;//1/x the hard way
   cout << "4:" << (sqf/cubf)[2.] << endl;
   z = &(expf*expf);
   cout << "5:" << (*z)(2.3025851)       << endl;
   cout << "6:" << (*z)[2.3025851]       << endl;
/* C1_function z;
   z = poly;
   cout << "5:" << z(3) << endl;
   cout << "6:" << z[3] << endl;  */
}

void ex_9_4(void)
{  double qpi = 2.*atan(1.0)/3.; //(30deg)
   cout << "1:" << (arg + sqf + cubf)(3.) << endl;
   cout << "2:" << (arg + sqf + cubf)[3.] << endl;
   Variable x;
   cout << "3:" << (x + sqf + cubf)(3.) << endl;
   cout << "4:" << (x + sqf + cubf)[3.] << endl;
   cout << "1:" << (1.0 - (2.0*(sinf*sinf)))(qpi) << endl; //cos(60)
   cout << "2:" << (1.0 - (2.0*(sinf*sinf)))[qpi] << endl;
   Parm one (1.0);
   cout << "0:" << (one + logf)[0.99] << endl;
   cout << "1:" << (1.0 + logf)[0.99] << endl;
}

void ex_9_5(void)
{
   double qpi = 2.*atan(1.0)/3.; //(30deg)
   C1_function bigfun, gradient[3], jacobian[3][3];
   C1_function cos2x( 1.0 - 2.0*sinf*sinf );
   bigfun = cos2x;  gradient[0] = sqf;  jacobian[2][2] = (expf*expf);
   cout << "1:" << cos2x(qpi) << endl;
   cout << "2:" << cos2x[qpi] << endl;
   cout << "3:" << bigfun(qpi) << endl;
   cout << "4:" << bigfun[qpi] << endl;
   cout << "5:" << gradient[0](-3.) << endl;
   cout << "6:" << gradient[0][-3.] << endl;
   cout << "7:" << jacobian[2][2](2.3025851) << endl;
   cout << "8:" << jacobian[2][2][2.3025851] << endl;
}

void ex_9_5_1(void)
{
   C1_function cosec( 1./sinf );
   C1_function   sec( 1./cosf );
   C1_function sin2x( 2.*sinf*cosf );
   C1_function cos3x( 4.*cosf*cosf*cosf - 3.*cosf );
   C1_function tan2x( 2.*tanf/((1.-tanf)*(1.+tanf)) );
   double qpi = 2.*atan(1.0)/3.; //(30deg)
   cout << "1:" << cosec(qpi) << endl;//1/sin(30)
   cout << "2:" << cosec[qpi] << endl;//-cos(30)/sin(30)/sin30
   cout << "3:" <<   sec(qpi) << endl;//1/cos(30)
   cout << "4:" <<   sec[qpi] << endl;//sin(30)/cos(30)/cos(30)
   cout << "5:" << sin2x(qpi) << endl;//sin(2*30)
   cout << "6:" << sin2x[qpi] << endl;//2*cos(2*30)
   cout << "7:" << cos3x(qpi) << endl;//cos(3*30)
   cout << "8:" << cos3x[qpi] << endl;//-3*sin(3*30)
   cout << "9:" << tan2x(qpi) << endl;//tan(2*30)
   cout << "10:" << tan2x[qpi] << endl;//2/cos(2*30)/cos(2*30)
   C1_function z; z = tan2x;
   cout << "11:" << z(qpi) << endl;
   cout << "12:" << z[qpi] << endl;
}

void ex_9_6(void)
{
   Variable x;
   Parm three(3.0), one(1.0), a;  a = 1.0;
   Chain A(Sum(x,one));    A.setnote("x+1");
   Chain B(logf);          B.setnote("ln");
   Chain C(Prod(x,three)); C.setnote("x*3");
   Chain D(Quot(a,sqf));   D.setnote("a/x**2");
   A >> B >> C;               // 3*ln(x+1)
   cout << A << B << C << D;
   cout << x(.5) << " | "
        << A(.5) << " | "
        << A[.5] << endl;
   cout << x(.5) << " | "
        << D(.5) << " | "
        << D[.5] << endl;
   cout << x(.5)     << " | "
        << (A+D)(.5) << " | "
        << (A+D)[.5] << endl;
}

void ex_9_7(void)
{
   Variable t;
   Parm x0, y0, z0, xs, ys, xr, yr, cx, cy, cz;
   C1_function x1sq( (x0+cx*t-xs)*(x0+cx*t-xs) );
   C1_function y1sq( (y0+cy*t-ys)*(y0+cy*t-ys) );
   C1_function z1sq( (z0+cz*t   )*(z0+cz*t   ) );
   Chain S( x1sq + y1sq + z1sq , "sumsq");
   Chain SSQRT( sqrtf, "sqrt()");
   S >> SSQRT;
   C1_function x2sq( (x0+cx*t-xr)*(x0+cx*t-xr) );
   C1_function y2sq( (y0+cy*t-yr)*(y0+cy*t-yr) );
   C1_function z2sq( (z0+cz*t   )*(z0+cz*t   ) );
   Chain R( x2sq + y2sq + z2sq , "sumsq");
   Chain RSQRT( sqrtf, "sqrt()");
   R >> RSQRT;
   x0 = 0.; y0 = 0.; z0 = 2.; xs = 1.; ys = 0.; xr = 2.; yr = 0.;
   cx = .6124; cy = .6124; cz = .5;
   for(double u=-1.; u<=1.; u+=.1)
      cout << t(u) << " | "
           << (S + R)(u) << " | "
           << (S + R)[u] << endl;
   WaitForIt();
   double u, du, w, dw, del, tol = 1.e-12;
   u = 1.;  du = (S + R)[u]; del = 0.1;
   for(int k=0; k<10; k++){
      cout << t(u) << " | " << (S + R)(u) << " | " << du << endl;
      if(fabs(del) < tol) break;
      w = u; dw = du; u += del; du = (S + R)[u];
      del = du*(w-u)/(du-dw);
      WaitForIt();
   }
/* Chain A(Sum(x,one));    A.setnote("x+1");
   Chain B(logf);          B.setnote("ln");
   Chain C(Prod(x,three)); C.setnote("x*3");
   Chain D(Quot(a,sqf));   D.setnote("a/x**2");
   A >> B >> C;               // 3*ln(x+1)
   cout << A << B << C << D;
   cout << x(.5) << " | "
        << A(.5) << " | "
        << A[.5] << endl;
   cout << x(.5) << " | "
        << D(.5) << " | "
        << D[.5] << endl;
   cout << x(.5)     << " | "
        << (A+D)(.5) << " | "
        << (A+D)[.5] << endl;   */
}

void main()
{
// ex_9_1_1();
// ex_9_3_1();
// ex_9_3_2();
// ex_9_4();
// ex_9_5_1();
// ex_9_6();
   ex_9_7();
   return;
}

